// ExpatXMLParser
// Wrapper class to translate from Objective-C messages, to proper expat C function calls.
// In addition to giving Cocoa programmers the ability to easily use the expat parser,
//  this class tries to conform to the NSXMLParser api specification.  This means that
//  expat can be a drop in replacement for the NSXMLParser, and in most cases, programmers
//  will only be required to change one line of code.  ('NSXMLParser' -> 'ExpatXMLParser')
//
// Written by Robbie Hanson
// Inspired by Rafael R. Sevilla's work on Expatobjc

#import "ExpatXMLParser.h"

// Default buffer size for parsing file
// File will be read from disk in chunks of X bytes
int BUFFER_SIZE = 8192;

// If parsing (processing) namespaces, expat returns elementNames
// as a concatenation of the namespaceURI + SEPARATOR + elementName.  (Same for attributes)
// Thus, this SEPARATOR character shouldn't be part of a legal URI.
// This character is the "registered trademark" character: Option-r on a mac keyboard
// 
// This line causes a warning to appear during compilation.
// If you can think of a character to go here that doesn't cause a warning, and isn't part of a legal URI
//  then you should place it here.  Then you should let the author know about it so he can change it too!
XML_Char SEPARATOR = '®';


@implementation ExpatXMLParser

// WRAPPER METHODS
//
// Wrapper functions that are used by Expat.  These are the "real"
// Expat handlers, and they receive the id of the object instance that
// is performing the parsing as their user data argument so they know
// where to send messages.
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// Called when an element (tag) is started
static void XMLCALL
startElementHandler(void *ud, const XML_Char *name, const XML_Char **atts)
{
	id parserobj = (id)ud;
	NSString *elementName;
	NSString *uri = nil;
	NSString *qualifiedName = nil;
	
	// Extract names
	if([parserobj shouldProcessNamespaces])
	{
		NSString *_name = [NSString stringWithCString:name];
		NSArray *temp = [_name componentsSeparatedByString:[NSString stringWithFormat:@"%c", SEPARATOR]];
		
		if([temp count] > 1)
		{
			uri = [temp objectAtIndex:0];
			elementName = [temp objectAtIndex:1];
		}
		else
			elementName = [temp objectAtIndex:0];
		
		// It's possible to have a default namespace, in which case a prefix may not have been declared
		if([temp count] > 2)
			qualifiedName = [NSString stringWithFormat:@"%@:%@", [temp objectAtIndex:2], elementName];
		else
			qualifiedName = elementName;
	}
	else
	{
		elementName = [NSString stringWithCString:name];
	}
	
	// Create attributes dictionary
	NSMutableDictionary *dict = [NSMutableDictionary dictionaryWithCapacity:1];
	int i;
	for(i=0; atts[i]; i+=2)
		[dict setObject:[NSString stringWithCString:atts[i+1]] forKey:[NSString stringWithCString:atts[i]]];
	
	[[parserobj delegate] parser:parserobj
				 didStartElement:elementName
					namespaceURI:uri
				   qualifiedName:qualifiedName
					  attributes:dict];
}

// Called when an element (tag) ends
static void XMLCALL
endElementHandler(void *ud, const XML_Char *name)
{
	id parserobj = (id)ud;
	[[parserobj delegate] parser:parserobj
				   didEndElement:[NSString stringWithCString:name]
				    namespaceURI:nil
				   qualifiedName:nil];
}

// Called when characters are encounted between elements (tags)
static void XMLCALL
characterDataHandler(void *ud, const XML_Char *s, int len)
{
	id parserobj = (id)ud;
	[[parserobj delegate] parser:parserobj foundCharacters:[NSString stringWithCString:s length:len]];
}

// Called when comments are encountered
static void XMLCALL
commentHandler(void *ud, const XML_Char *data)
{
	NSString *_comment = (data == NULL) ? nil : [NSString stringWithCString:data];
	
	id parserobj = (id)ud;
	[[parserobj delegate] parser:parserobj foundComment:_comment];
}

// Called when a namespace is declared
static void XMLCALL
startNamespaceDeclHandler(void *ud, const XML_Char *prefix, const XML_Char *uri)
{
	NSString *_prefix = (prefix == NULL) ? @"" : [NSString stringWithCString:prefix];
	NSString *_uri = (uri == NULL) ? nil : [NSString stringWithCString:uri];
	
	id parserobj = (id)ud;
	[[parserobj delegate] parser:parserobj didStartMappingPrefix:_prefix toURI:_uri];
}

// Called when a namespace declaration ends (falls out of scope)
static void XMLCALL
endNamespaceDeclHandler(void *ud, const XML_Char *prefix)
{
	NSString *_prefix = (prefix == NULL) ? @"" : [NSString stringWithCString:prefix];
	
	id parserobj = (id)ud;
	[[parserobj delegate] parser:parserobj didEndMappingPrefix:_prefix];
}

static void XMLCALL
processingInstructionHandler(void *ud, const XML_Char *target, const XML_Char *data)
{
	NSString *_target = (target == NULL) ? nil : [NSString stringWithCString:target];
	NSString *_data = (data == NULL) ? nil : [NSString stringWithCString:data];
	
	id parserobj = (id)ud;
	[[parserobj delegate] parser:parserobj foundProcessingInstructionWithTarget:_target data:_data];
}

// OBJECTIVE-C METHODS
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

- (id)initWithContentsOfURL:(NSURL *)url
{
	return [self initWithData:[NSData dataWithContentsOfURL:url]];
}

- (id)initWithData:(NSData *)new_data
{
	if(self = [super init])
	{
		shouldProcessNamespaces = NO;
		shouldReportNamespacePrefixes = NO;
		
		data = [new_data copy];
	}
	return self;
}

-(void)dealloc
{
	XML_ParserFree(parser);
	[data release];
	[error release];
	[super dealloc];
}

-(id)delegate
{
	return delegate;
}

-(void)setDelegate:(id)newDelegate
{
	delegate = newDelegate;
}

- (void)setShouldProcessNamespaces:(BOOL)flag
{
	shouldProcessNamespaces = flag;
}

- (BOOL)shouldProcessNamespaces
{
	return shouldProcessNamespaces;
}

- (void)setShouldReportNamespacePrefixes:(BOOL)flag
{
	shouldReportNamespacePrefixes = flag;
}

- (BOOL)shouldReportNamespacePrefixes
{
	return shouldReportNamespacePrefixes;
}

- (void)setShouldResolveExternalEntities:(BOOL)flag
{
	shouldResolveExternalEntities = flag;
}

- (BOOL)shouldResolveExternalEntities
{
	return shouldResolveExternalEntities;
}

- (void)errorHandler:(int)errorCode
{
	if([delegate respondsToSelector:@selector(parser:parseErrorOccurred:)])
	{
		NSString *domain = @"ExpatXMLParserDomain";
		NSString *errorStr = [NSString stringWithCString:XML_ErrorString(errorCode)];
		
		NSDictionary *errorInfo = [NSDictionary dictionaryWithObject:errorStr forKey:NSLocalizedDescriptionKey];
		
		error = [[NSError alloc] initWithDomain:domain code:errorCode userInfo:errorInfo];
		
		[delegate parser:self parseErrorOccurred:error];
	}
}

- (BOOL)parse
{	
	// Initialize parser
	if(shouldProcessNamespaces)
	{
		parser = XML_ParserCreateNS(NULL,SEPARATOR);
		XML_SetReturnNSTriplet(parser,1);
	}
	else
		parser = XML_ParserCreate(NULL);
	
	// Configure parser to pass proper user data
	XML_SetUserData(parser, self);
	
	// Set Standard Handlers
	if([delegate respondsToSelector:@selector(parser:didStartElement:namespaceURI:qualifiedName:attributes:)])
		XML_SetStartElementHandler(parser, startElementHandler);
	
	if([delegate respondsToSelector:@selector(parser:didEndElement:namespaceURI:qualifiedName:)])
		XML_SetEndElementHandler(parser, endElementHandler);
	
	if([delegate respondsToSelector:@selector(parser:foundCharacters:)])
		XML_SetCharacterDataHandler(parser, characterDataHandler);
	
	if([delegate respondsToSelector:@selector(parser:foundComment:)])
		XML_SetCommentHandler(parser, commentHandler);
	
	if([delegate respondsToSelector:@selector(parser:foundProcessingInstructionWithTarget:data:)])
		XML_SetProcessingInstructionHandler(parser, processingInstructionHandler);
	
	// Set Namespace Handlers
	if(shouldReportNamespacePrefixes)
	{
		if([delegate respondsToSelector:@selector(parser:didStartMappingPrefix:toURI:)])
			XML_SetStartNamespaceDeclHandler(parser, startNamespaceDeclHandler);
		
		if([delegate respondsToSelector:@selector(parser:didEndMappingPrefix:)])
			XML_SetEndNamespaceDeclHandler(parser, endNamespaceDeclHandler);
	}
	
	// Notify of starting document
	if([delegate respondsToSelector:@selector(parserDidStartDocument:)])
		[delegate parserDidStartDocument:self];
	
	NSRange range;
	range.location = 0;
	
	BOOL done = NO;
	while(!done)
	{
		void *buffer = XML_GetBuffer(parser, BUFFER_SIZE);
		
		if(buffer == NULL)
		{
			[self errorHandler:XML_GetErrorCode(parser)];
			return NO;
		}
		else
		{
			// Get range of bytes to read
			if([data length] < BUFFER_SIZE + range.location)
			{
				range.length = [data length] - range.location;
				done = YES;
			}
			else
				range.length = BUFFER_SIZE;
			
			// Copy range into buffer
			[data getBytes:buffer range:range];
			
			// Parse bytes in buffer
			if(!XML_ParseBuffer(parser, range.length, done))
			{
				[self errorHandler:XML_GetErrorCode(parser)];
				return NO;
			}
			
			// Update range
			range.location += range.length;
		}
	}
	
	// Notify of ending document
	if([delegate respondsToSelector:@selector(parserDidEndDocument:)])
		[delegate parserDidEndDocument:self];
	
	return YES;
}

- (void)abortParsing
{
	XML_StopParser(parser, XML_FALSE);
}

- (NSError *)parserError
{
	return error;
}

- (int)columnNumber
{
	// NSXMLParser returns column number of the end of the tag
	// Expat returns column number of the beginning of the tag,
	//  so we add the length of the tag
	if(parser == NULL)
		return 0;
	else
		return XML_GetCurrentColumnNumber(parser) + XML_GetCurrentByteCount(parser);
}

- (int)lineNumber
{
	if(parser == NULL)
		return 0;
	else
		return XML_GetCurrentLineNumber(parser);
}

- (NSString *)publicID
{
	return nil;
}

- (NSString *)systemID
{
	return nil;
}

@end
